"use strict";
var __importStar = (this && this.__importStar) || function (mod) {
    if (mod && mod.__esModule) return mod;
    var result = {};
    if (mod != null) for (var k in mod) if (Object.hasOwnProperty.call(mod, k)) result[k] = mod[k];
    result["default"] = mod;
    return result;
};
var __importDefault = (this && this.__importDefault) || function (mod) {
    return (mod && mod.__esModule) ? mod : { "default": mod };
};
Object.defineProperty(exports, "__esModule", { value: true });
var t = __importStar(require("@babel/types"));
var traverse_1 = __importDefault(require("@babel/traverse"));
/**
 * Returns the binding identifier for a member expression.
 * For example the member expression `foo.bar.baz` will return the `foo` identifier.
 *
 * @param expression - Member expression node.
 */
exports.getMemberExpressionMeta = function (expression) {
    var bindingIdentifier = null;
    var accessPath = [];
    if (t.isIdentifier(expression.property)) {
        accessPath.push(expression.property);
    }
    traverse_1.default(t.expressionStatement(expression), {
        noScope: true,
        MemberExpression: function (path) {
            if (t.isIdentifier(path.node.object)) {
                bindingIdentifier = path.node.object;
            }
            if (t.isIdentifier(path.node.property)) {
                accessPath.push(path.node.property);
            }
        },
    });
    return {
        bindingIdentifier: bindingIdentifier,
        accessPath: accessPath.reverse(),
    };
};
/**
 * Will return the value of a path from an object expression.
 *
 * For example if  we take an object expression that looks like:
 * ```
 * { colors: { primary: 'red' } }
 * ```
 *
 * And a path of identifiers that looks like:
 * ```
 * [colors, primary]
 * ```
 *
 * Would result in returning the `red` string literal node.
 * If the value is not found `undefined` will be returned.
 *
 * @param expression - Member expression node.
 * @param accessPath - Access path identifiers.
 */
exports.getValueFromObjectExpression = function (expression, accessPath) {
    var value = undefined;
    traverse_1.default(expression, {
        noScope: true,
        ObjectProperty: function (path) {
            if (t.isIdentifier(path.node.key, { name: accessPath[0].name })) {
                if (t.isObjectExpression(path.node.value)) {
                    value = exports.getValueFromObjectExpression(path.node.value, accessPath.slice(1));
                }
                else {
                    value = path.node.value;
                }
                path.stop();
            }
        },
    });
    return value;
};
/**
 * Will return either the name of an identifier or the value of a string literal.
 *
 * E.g:
 * - `foo` identifier node will return `"foo"`,
 * - `"bar"` string literal node will return `"bar"`.
 *
 * @param node
 */
exports.getKey = function (node) {
    if (t.isIdentifier(node)) {
        return node.name;
    }
    if (t.isStringLiteral(node)) {
        return node.value;
    }
    throw new Error();
};
/**
 * Will look in an expression and return the actual value.
 * If the expression is an identifier node (a variable) and a constant,
 * it will return the variable reference.
 *
 * E.g: If there was a identifier called `color` that is set somewhere as `const color = 'blue'`,
 * passing the `color` identifier to this function would return `'blue'`.
 *
 * This behaviour is the same for const string & numeric literals,
 * and object expressions.
 *
 * @param expression Expression we want to interrogate.
 * @param state Babel state - should house options and meta data used during the transformation.
 */
exports.getInterpolation = function (expression, meta) {
    var value = undefined;
    if (t.isIdentifier(expression)) {
        var binding = meta.parentPath.scope.getBinding(expression.name);
        if (binding && binding.kind === 'const' && t.isVariableDeclarator(binding.path.node)) {
            value = binding.path.node.init;
        }
    }
    else if (t.isMemberExpression(expression)) {
        var _a = exports.getMemberExpressionMeta(expression), accessPath = _a.accessPath, bindingIdentifier = _a.bindingIdentifier;
        var binding = meta.parentPath.scope.getBinding(bindingIdentifier.name);
        if (binding &&
            binding.kind === 'const' &&
            t.isVariableDeclarator(binding.path.node) &&
            t.isObjectExpression(binding.path.node.init)) {
            value = exports.getValueFromObjectExpression(binding.path.node.init, accessPath);
        }
    }
    if (t.isStringLiteral(value) || t.isNumericLiteral(value) || t.isObjectExpression(value)) {
        return value;
    }
    return expression;
};
/**
 * Will wrap BlockStatement or Expression in an IIFE,
 * Looks like (() => { return 10; })().
 *
 * @param node Node of type either BlockStatement or Expression
 */
exports.wrapNodeInIIFE = function (node) {
    return t.callExpression(t.arrowFunctionExpression([], node), []);
};
var tryWrappingBlockStatementInIIFE = function (node) {
    return t.isBlockStatement(node) ? exports.wrapNodeInIIFE(node) : node;
};
/**
 * Will pick `ArrowFunctionExpression` body and tries to wrap it in an IIFE if
 * its a BlockStatement otherwise returns the picked body,
 * E.g.
 * `props => props.color` would end up as `props.color`.
 * `props => { return props.color` } would end up as `(() => { return props.color })()`.
 *
 * @param node Node of type ArrowFunctionExpression
 */
exports.pickArrowFunctionExpressionBody = function (node) {
    return tryWrappingBlockStatementInIIFE(node.body);
};
//# sourceMappingURL=ast.js.map